
import sys
import os
import itertools
import operator

from qgis.PyQt.QtCore import QVariant
from qgis.PyQt.QtGui import QColor
from qgis.PyQt.QtCore import Qt
from qgis.core import *

# Translation from gas SHapeCodes to a QGIS symbol - a mixed list of built-in symbolMarker definitions, and font charaters
QGIS_SYMBOLS =    [('MARKER', QgsSimpleMarkerSymbolLayerBase.Circle, 0, True),  # 0-8 filled
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Square, 0, True),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, 180, True),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, 0, True),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Diamond, 0, True),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.SemiCircle, 270, True),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.SemiCircle, 90, True),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.HalfSquare, 0, True),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.HalfSquare, 90, True),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Circle, 0, False),  # 9-17 Unfilled
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Square, 0, False),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, 180, False),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, 0, False),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Diamond, 0, False),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.SemiCircle, 270, False),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.SemiCircle, 90, False),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.HalfSquare, 0, False),
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.HalfSquare, 90, False),

                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Cross2, 90, False),  # X  #18
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Cross, 90, False),  # +  #19

                   # These font based markers apparently not responding to colour in symbology
                   ('FONT','Arial', 'C', 0, False),  #20 C
                   ('FONT','Arial', 'U', 0, False),  # U
                   ('FONT','Arial', '<', 0, False),  # LT <
                   ('FONT','Arial', 'V', 0, False),  # V
                   ('FONT','Arial', 'L', 90, False),  # L rotated gamma
                   ('FONT','Arial', 'L', 0, False),  #25 L  #

                   ('FONT','Arial', 'Z', 0, False),  #26  Z
                   ('FONT','Arial', 'N', 0, False),  #  N
                   ('FONT','Arial', '/', 0, False),  #  /
                   ('FONT','Arial', '\\', 0, False),  #29 \

                   ('FONT','Wingdings', '\xE0', 0, False),  #30  224 r-arrow #30
                   ('FONT','Wingdings', '\xE1', 0, False),  # 225 up-arrow
                   ('FONT','Arial', 'Y', 180, False),  # orig wingdings 229 3 way thingy upside down Y
                   ('FONT','Wingdings', '\xAB', 0, False),  # 171 star
                   ('FONT','Arial', '8', 90, False),  # infinity  8 rotated
                   ('FONT','Arial', '8', 0, False),  #35   char 8
                   ('FONT','Arial', '\xB6', 90, False),  # pilcrow (CR symbol)  8 rotated
                   ('FONT','Arial', '\xB6', 0, False),  #37 pilcrow


                   # ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, -45, False),  #38  quadrant triangles38
                   # ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, -135, False),
                   # ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, -45, True),
                   # ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, -135, True),  #41

                    # these 4 shapes changed june 2019 JP
                   ('MARKER', QgsSimpleMarkerSymbolLayerBase.DiagonalHalfSquare, 90, False),  # 38  DiagonalHalfSquare pointing Top Left, UNfilled
                   ('MARKER', QgsSimpleMarkerSymbolLayerBase.DiagonalHalfSquare, 0, False),  # 39  DiagonalHalfSquare pointing Bottom Left, UNfilled
                   ('MARKER', QgsSimpleMarkerSymbolLayerBase.DiagonalHalfSquare, 90, True),  # 40  DiagonalHalfSquare pointing Top Left, Filled
                   ('MARKER', QgsSimpleMarkerSymbolLayerBase.DiagonalHalfSquare, 0, True),  # 41  DiagonalHalfSquare pointing Bottom Left, Filled

                   # new symbols june 2019
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, -90, False),  #42 left pointing arrowhead unfilled
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, 90, False),  # right pointing arrowhead unfilled
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, -90, True),  # left pointing arrowhead filled
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.Triangle, 90, True),  # right pointing arrowhead filled
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.DiagonalHalfSquare, 180, False),  # DiagonalHalfSquare pointing Top Right, unfilled
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.DiagonalHalfSquare, -90, False),  # DiagonalHalfSquare pointing Bottom Right. unfilled
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.DiagonalHalfSquare, 180, True),  # DiagonalHalfSquare pointing Top Right, Filled
                   ('MARKER',QgsSimpleMarkerSymbolLayerBase.DiagonalHalfSquare, -90, True)  #49 DiagonalHalfSquare pointing Bottom Right, Filled

                   ]

class GasLogic:
    LEGEND_LIMIT = 60  # >1 if we need to limit number of legend entries - only used where qgis<3.3

    def __init__(self):
        self.r = None
        self.outputFile = None
        self.legendType = ''

    def setOutputFile(self, outputFile):
        self.outputFile = outputFile

    def setReader(self, r):
        self.r = r

    def setLegendType(self, legendType):
        self.legendType = legendType

    def createGeoPackageLayer(self, outputFile,layerName):
        self.outputFile = outputFile
        layer = self.createGeoPackageFile(layerName)
        return layer

    # convert gasfilereader object to a geopackage file on the filesystem
    def createGeoPackageFile(self,layerName):

        self.removeLayerIfExists(layerName)

        QgsMessageLog.logMessage("legendType:" + self.legendType, 'ioGAS', level=Qgis.Info)

        # coord sys
        epsg = self.r.epsg
        QgsMessageLog.logMessage('setting crs ' + str(epsg), 'ioGAS', level=Qgis.Info)
        crs = QgsCoordinateReferenceSystem('EPSG:' + epsg)


        # define a feature - a template if you like
        ftr = QgsFeature()
        fields = self.setupFieldStructure()
        ftr.setFields(fields)

        # create a new empty point feature layer at the output file location, with the desired attribute structure
        writer = QgsVectorFileWriter(self.outputFile, 'utf-8', fields, QgsWkbTypes.Point, crs)

        # for each row in source data (ie for each source point)
        for row in self.r.dataAndAtts:
            #QgsMessageLog.logMessage("row  " + str(row) , 'ioGAS', level=Qgis.Info)

            # add data values to attribute collection
            self.setFeatureAttributeValuesFromRow(row,ftr)

            self.setFeatureGeometryFromRow(row,ftr)

            # add feature to file
            writer.addFeature(ftr)

        del writer # done writing features

        # create layer from geopackage file and apply the symbology to the layer
        layer = QgsVectorLayer(self.outputFile, layerName, 'ogr')

        self.setLayerSorting(layer)

        QgsMessageLog.logMessage("Layer Created from Geopackage file", 'ioGAS',level=Qgis.Info)

        self.applySymbologyToLayer(layer, True)

        return layer


    # Create an in-memory layer (aka "Scratch" layer)
    def createMemoryLayer(self,layerName):

        self.removeLayerIfExists(layerName)

        QgsMessageLog.logMessage("Writing Memory Vector Layer : legendType:" + self.legendType, 'ioGAS', level=Qgis.Info)

        # coord sys
        epsg = self.r.epsg
        # crs = QgsCoordinateReferenceSystem('EPSG:' + epsg)   # currently we explicitly set mem layer crs  using epsg code -
        QgsMessageLog.logMessage('setting crs ' + str(epsg), 'ioGAS', level=Qgis.Info)

        # define a feature - a kind of template
        ftr = QgsFeature()
        fields = self.setupFieldStructure()
        ftr.setFields(fields)

        # create empty layer
        layer = QgsVectorLayer(f"Point?crs=EPSG:{epsg}", layerName, "memory")
        self.setLayerSorting(layer)

        # configure data provider for layer with attribute structure
        provider = layer.dataProvider()
        provider.addAttributes(fields)

        # changes are only possible when editing the layer
        layer.startEditing()

        # for each row in source data (ie for each source point)
        for row in self.r.dataAndAtts:
            # QgsMessageLog.logMessage(("s row  " + str(row) + " x, y= " + str(x) + ", " + str(y)), 'ioGAS')

            # add data values to attribute collection
            self.setFeatureAttributeValuesFromRow(row,ftr)

            self.setFeatureGeometryFromRow(row, ftr)

            provider.addFeatures([ftr]) # add feature to layer via it's data provider

        layer.commitChanges() # stop editing and save features to layer

        layer.updateExtents() # update layer's spatia extent after new features have been added

        self.applySymbologyToLayer(layer,False)

        return layer

        # apply the symbology to the layer

    def applySymbologyToLayer(self,layer,isGeoPackage):
        rule = None
        # if self.legendType == 'HIERARCHICAL':
        #     rule = self.generateRulesHierarchical()
        # else:
        #     rule = self.generateRulesLinear()

        rule = self.generateRulesLinear()
        renderer = QgsRuleBasedRenderer(rule)
        layer.setRenderer(renderer)
        if isGeoPackage: layer.saveStyleToDatabase('main', 'ioGAS', True, '')  # for gpkg we save the style into the gpkg file


    #Add required empty column structure to hold attribute values that will be applied to each created point feature
    def setupFieldStructure(self):
        fields = QgsFields()
        for i, c in enumerate(self.r.columns):
            dataType = QVariant.String
            if c.type == "Text":
                dataType = QVariant.String
            if c.type == "Numeric":
                dataType = QVariant.Double

            f = QgsField(c.name, dataType)
            fields.append(f)

        # gas special attribute fields - we need these as they drive the legend via the legend rules
        fields.append(QgsField(self.r.GAS_FILTER, QVariant.String))
        fields.append(QgsField(self.r.GAS_COLOUR, QVariant.String))
        fields.append(QgsField(self.r.GAS_SHAPE, QVariant.String))
        fields.append(QgsField(self.r.GAS_SIZE, QVariant.String))

        return fields


    # set feature attribute values from data row
    # NB believe Qgis will cast numerics as required using info in the features fields collection
    def setFeatureAttributeValuesFromRow(self,row,ftr):
        attributes = []
        for v in row:
            if len(v) == 0:  # add as null
                attributes.append(None)
            else:
                if v.isnumeric():  # this cast is prob unnecessary but we keep it in case we need to do more explicit casting in future
                    attributes.append(float(v))
                else:
                    attributes.append(v)

        ftr.setAttributes(attributes)

    # set feature geometry/location from coord values in the data row
    def setFeatureGeometryFromRow(self, row, ftr):
        # set feature location
        if self.r.indexZ == -1:  # 2D
            try:
                x = float(row[self.r.indexX])  # coords
                y = float(row[self.r.indexY])
                ftr.setGeometry(QgsGeometry.fromPointXY(QgsPointXY(x, y)))
            except:
                QgsMessageLog.logMessage("row with bad x,y", 'ioGAS', level=Qgis.Warning)

        else:  # 3D
            try:
                x = float(row[self.r.indexX])  # coords
                y = float(row[self.r.indexY])
                z = float(row[self.r.indexZ])
                ftr.setGeometry(QgsGeometry(QgsPoint(x, y, z)))
            except:
                print("Unexpected Geometry error:", sys.exc_info()[0])
                QgsMessageLog.logMessage("row with bad x,y,z", 'ioGAS', level=Qgis.Warning)


    # composite (linear) renderer  - creates a rule for each unique combination of attributes (Colour & Shape & Size) (as per esri)
    def generateRulesLinear(self):
        ruleTemplate = '"{field1}" = \'{value1}\' and  "{field2}" = \'{value2}\' and  "{field3}" = \'{value3}\'  '
        rootRule = QgsRuleBasedRenderer.Rule(None)
        QgsMessageLog.logMessage("Making rules: " + str(len(self.r.uniqueAttCombinations)), "ioGAS", level=Qgis.Info)

        for uaIndex, ua in enumerate(self.r.uniqueAttCombinations):
            #indices into shape/colour/size atts for this unique combination
            colourAtt = int(ua[1])
            colour = self.r.colours[colourAtt]  # entry from gas metadata colour lookup table
            shapeAtt = int(ua[2])
            shape = self.r.shapes[shapeAtt] # entry from gas metadata shape lookup table
            sizeAtt = int(ua[3])
            size = self.r.sizes[sizeAtt] # entry from gas metadata size lookup table

            #Create a symbol and related rule the rule
            symbol = QgsSymbol.defaultSymbol(QgsWkbTypes.PointGeometry).clone()
            rule = QgsRuleBasedRenderer.Rule(symbol)

            # Style the symbol
            qgisSymbolRef = QGIS_SYMBOLS[int(shape.shapeCode) % len(   QGIS_SYMBOLS)]  # a tuple from our symbol Library  MAP_SYMBOLS, wrapping where index higher than number of items in list

            self.SetSymbolShapeAndColour(symbol, shape, colour)

            self.SetSymbolSize(symbol, size)

            # Configure the rule - we get the above symbol if the rule matches (ie if the rows shape/colour/size att values all match)
            rule.setFilterExpression(ruleTemplate.format(field1=self.r.GAS_COLOUR, value1=colourAtt, field2=self.r.GAS_SHAPE, value2=shapeAtt, field3=self.r.GAS_SIZE, value3=sizeAtt))
            rule.setLabel(colour.name + ", " + shape.name + ", " + size.name)
            rule.setDescription("ioGAS rule [" + str(uaIndex) + "] " + str(ua))
            # QgsMessageLog.logMessage( expr + " indexes " + str(colourAtt) + ' ' + str(shapeAttr) + ' ' +  str(sizeAtt) + "  colour : " + str(colour), 'ioGAS')

            rootRule.appendChild(rule)

        return rootRule

    # Create a rule-based renderer that will implement a hierachy of colour/shape/size
    # def generateRulesHierarchical(self):
    #
    #     # grab the defaults
    #     defaultColour = self.r.colours[0]
    #     defaultShape = self.r.shapes[0]
    #     defaultSize = self.r.sizes[0]
    #
    #     ruleTemplate = '"{field}" = \'{value}\''  # generic template rule def that will match a single field/value
    #     rootRule = QgsRuleBasedRenderer.Rule(None)
    #
    #
    #     # unique combinations tuple :  # 0 special, 1 colour, 2 shape, 3 size
    #
    #     # Build a list of Unique colours used - by extracting just the colour attribute list, convert to a set and then back to a list
    #     usedColourAtts = list(set([x[1] for x in self.r.uniqueAttCombinations]))
    #
    #     # for each unique colour used
    #     for colourAtt in usedColourAtts:
    #
    #         colour = self.r.colours[colourAtt]  #  get gasfile lookup table entry using the colour Attribute
    #
    #         # Create and Colour ua symbol for this level in hierarchy
    #         colourRuleSymbol = QgsSymbol.defaultSymbol(QgsWkbTypes.PointGeometry)
    #         self.SetSymbolColour(colourRuleSymbol, colour)   # Use current colour, default shape and size
    #         self.SetSymbolShapeAndColour(colourRuleSymbol, defaultShape)
    #         self.SetSymbolSize(colourRuleSymbol, defaultSize)
    #
    #
    #         # Create the Rule
    #         colourRule = QgsRuleBasedRenderer.Rule(colourRuleSymbol) # option1 - we get a top level icon
    #         colourRule.setFilterExpression(ruleTemplate.format(field=self.r.GAS_COLOUR, value=colourAtt))
    #         colourRule.setLabel(colour.name)
    #         colourRule.setDescription(colour.name)
    #
    #         # Build ua distinct list of usedShapes *for the current colour*
    #         usedShapeAtts = list(set([x[2] for x in self.r.uniqueAttCombinations if x[1] == colourAtt]))
    #
    #         # for each distinct shape for current colour, create a rule and symbol styled with shape, default size - will inherit current colour
    #         for shapeAtt in usedShapeAtts : # attrs in usedShapes.items():
    #
    #             shape = self.r.shapes[shapeAtt]   # get gasfile lookup table entry from the shape Attribute
    #
    #             # create & style the symbol  - has current colour and shape, but default size
    #             shapeRuleSymbol = QgsSymbol.defaultSymbol(QgsWkbTypes.PointGeometry)
    #
    #             self.SetSymbolColour(shapeRuleSymbol, colour)
    #             self.SetSymbolShapeAndColour(shapeRuleSymbol, shape)
    #             self.SetSymbolSize(shapeRuleSymbol, defaultSize)
    #
    #             # Create the Rule ;
    #             shapeRule = QgsRuleBasedRenderer.Rule(shapeRuleSymbol)
    #             shapeRule.setFilterExpression(ruleTemplate.format(field=self.r.GAS_SHAPE, value=shapeAtt))
    #             shapeRule.setLabel(colour.name + ", " + shape.name)
    #             shapeRule.setDescription(colour.name + ", " + shape.name)
    #
    #             # Build ua distinct list of Sizes for the current colour & shape
    #             usedSizeAtts = list(set([x[3] for x in self.r.uniqueAttCombinations if x[1] == colourAtt and x[2] == shapeAtt ]))
    #
    #             # For each distinct Size style ua symbol
    #             for sizeAtt in usedSizeAtts:#, attrs in sizes.items():
    #                 size = self.r.sizes[sizeAtt]  # get gasfile lookup table entry using this size attribute
    #
    #                 # create & style the symbol - now has current colour + shape + size
    #                 sizeRuleSymbol = QgsSymbol.defaultSymbol(QgsWkbTypes.PointGeometry)
    #                 self.SetSymbolColour(sizeRuleSymbol, colour)
    #                 self.SetSymbolShapeAndColour(sizeRuleSymbol, shape)
    #                 self.SetSymbolSize(sizeRuleSymbol, size)
    #
    #                 # Create the Rule specifying size, will inherit current colour and shape
    #                 sizeRule = QgsRuleBasedRenderer.Rule(sizeRuleSymbol)
    #                 sizeRule.setFilterExpression(ruleTemplate.format(field=self.r.GAS_SIZE, value=sizeAtt))
    #                 sizeRule.setLabel(colour.name + ", " + shape.name + ", " + size.name)
    #                 sizeRule.setDescription(colour.name + ", " + shape.name + ", " + size.name)
    #
    #                 # Building the hierachy of rules as we unwind
    #                 shapeRule.appendChild(sizeRule)  # Add size rule as child of shape rule
    #
    #             colourRule.appendChild(shapeRule) # Add Shape rule as child of Colour rule
    #
    #         rootRule.appendChild(colourRule)  # finally add to the root
    #
    #     return rootRule

    # # Resolve rgb colour from gas lookup table entry and apply to symbol
    # def SetSymbolColour(self, symbol, colour):
    #     rgbaQcolour = QColor.fromRgba(int(colour.colour)) #  convert java rgba colour from gas lookup table to QColor
    #     symbol.symbolLayer(0).setStrokeColor(rgbaQcolour)
    #     symbol.symbolLayer(0).setFillColor(rgbaQcolour)
    #     return rgbaQcolour
    #
    # def SetFontSymbolColour(self, symbol, colour):   # font based symbols clured using SetColor
    #     rgbaQcolour = QColor.fromRgba(int(colour.colour))  # convert java rgba colour from gas lookup table to QColor
    #     symbol.symbolLayer(0).setFillColor(rgbaQcolour)
    #     symbol.symbolLayer(0).setColor(rgbaQcolour)
    #     return rgbaQcolour

    #  Set symbol shape - look up a shape from our qgis shapes list and set for this symbol

    def SetSymbolShapeAndColour(self, symbol, shape, colour):

        # TODO : If symbol index is outside of dictionary range, display a black ! (bang)
        qgisSymbolRef = QGIS_SYMBOLS[int(shape.shapeCode) % len(QGIS_SYMBOLS)]    # a tuple from our symbol Library  MAP_SYMBOLS, wrapping where index higher than number of items in list
        rgbaQcolour = QColor.fromRgba(int(colour.colour))  # convert java rgba colour from gas lookup table to QColor

        if qgisSymbolRef[0] == 'FONT':  # font-based  ('FONT','Arial', 'C', 0, False),  # C  #20
            symbol.changeSymbolLayer(0, QgsFontMarkerSymbolLayer(qgisSymbolRef[1], qgisSymbolRef[2]))
            symbol.setAngle(qgisSymbolRef[3])
            symbol.symbolLayer(0).setColor(rgbaQcolour)
            # if qgisSymbolRef[4] != True:  # unfilled
            #     rgbaQcolourFill = QColor(0, 0, 0, 0)  # transparent
            #     symbol.symbolLayer(0).setColor(rgbaQcolourFill)
        else:  # Marker eg ('MARKER',QgsSimpleMarkerSymbolLayerBase.Circle, 0, True)
            symbol.symbolLayer(0).setShape(qgisSymbolRef[1])
            symbol.setAngle(qgisSymbolRef[2])
            symbol.symbolLayer(0).setStrokeColor(rgbaQcolour)
            if qgisSymbolRef[3] != True:  # unfilled
                rgbaTransparent = QColor(0, 0, 0, 0)  # transparent
                symbol.symbolLayer(0).setFillColor(rgbaTransparent)
            else:
                symbol.symbolLayer(0).setFillColor(rgbaQcolour)
        return


    # Set symbol size from from gas lookup table entry and apply to symbol
    def SetSymbolSize(self, symbol, size):
        pointSize = float(size.size) / 3.0  # hard coded scaling factor
        symbol.setSize(pointSize)
        return

    # Attempt to force attribute table viewer rows to be sorted in feature added order, for a layer
    def setLayerSorting(self, layer):
        config = layer.attributeTableConfig()
        config.setSortExpression('$id')  # inbuilt feature id ascends in order they were added
        config.setSortOrder(Qt.AscendingOrder)
        layer.setAttributeTableConfig(config)


    # return true if we are not able to display legend due to qgis version vs number of legend compbos
    # prior to qgis 3.3 (version int = 30300) there is a limit to how large a linear legend can be,
    # caused by an untrappable failure with very large sql statements in sqllite
    # result is that we return false for linear legends where qgis ver < 3.3 and number combos > LEGEND_LIMIT
    # NB: Finding version number may cause exception on earlier versions of qgis; but we really only care whether we have ver < 3.3
    def canDisplayLegend(self):

        if self.legendType!="LINEAR" :   # always display non-Linear Legend
            return True

        # do we have qgis < 3.3
        QGISVersion_LT3_3 = True
        try:
            if Qgis.QGIS_VERSION_INT  >= 30300 : # this will error on qgis versions <3 (?) leaving false
                QGISVersion_LT3_3 = False
        except:
            pass

        if QGISVersion_LT3_3 and len(self.r.uniqueAttCombinations) > self.LEGEND_LIMIT : return False  # can't display
        else: return True


    # Remove any map layer of same name
    # NB, we are not deleting files, just removing from the map's layers collection
    def removeLayerIfExists(self, layerName):
        layersall = QgsProject.instance().mapLayers()
        layers = QgsProject.instance().mapLayersByName(layerName)
        if len(layers) > 0:
            for layer in layers:
                QgsMessageLog.logMessage(f"Removing previous layer with this name ({layerName}).", 'ioGAS', level=Qgis.Warning)
                QgsProject.instance().removeMapLayers([layer.id()])  #QgsProject.instance().removeMapLayers( [vl.id()] )
