
import os
import itertools, operator

# allow for different spellings of Qgis/QGis between qgis versions
# try:
#     from qgis.core import Qgis # qgis ver 3 on
# except ImportError:
#     from qgis.core import QGis as Qgis  # qgis ver < 3

from qgis.core import *

#import argparse


#start of this path 'ioGAS' must match deployment
from gasfileloader import GasFileLoader
from gaslogic import GasLogic

from qgis.core import QgsApplication

buildID  = "104612"

msg = f"Loaded QGIS for ioGAS Plugin ver {buildID}"
print(msg)


def createGpkgLayer(src,legendType):
    print("Reading :" + src)
    # Read Gas file
    app = GasLogic()
    reader = GasFileLoader(src)
    app.setReader(reader)
    app.setLegendType(legendType)
    layerName = os.path.splitext(os.path.basename(src))[0]  # filename only w/out extension

    # check ua limit
    if not app.canDisplayLegend():
        print('Must be not more than ' + str(app.LEGEND_LIMIT) + " unique attributes  for Linear Legend")
        return

    dest = os.path.normpath(os.path.join(os.path.dirname(src), os.path.splitext(os.path.basename(src))[0] + ".gpkg"))
    print("Writing Geopackage :" + dest)
    layer = app.createGeoPackageLayer(dest,layerName)

def createScratchLayer(src,legendType):
    print("Reading :" + src)
    # Read Gas file
    app = GasLogic()
    reader = GasFileLoader(src)
    app.setReader(reader)
    app.setLegendType(legendType)
    layerName = os.path.splitext(os.path.basename(src))[0]  # filename only w/out extension
    layerName = layerName + " [Scratch]"

    # check ua limit
    if not app.canDisplayLegend():
        print('Must be not more than ' + str(app.LEGEND_LIMIT) + " unique attributes  for Linear Legend")
        return

    print("Writing Scratch Layer")

    layer = app.createMemoryLayer(layerName)


#currently unused
# uncomment below to run as a command line accepting file in/out paths as arguments
# if __name__ == '__main__':
#    parser = argparse.ArgumentParser(description='Convert GAS file to GeoPackage.')
#    parser.add_argument('gas_file', metavar='GAS_FILE', type=str, nargs=1, help='input GAS file')
#    parser.add_argument('out_file', metavar='OUT_FILE', type=str, nargs=1, help='output GeoPackage')
#    args = parser.parse_args()
#    print(args)

print("Initialising ")
QgsApplication.setPrefixPath('/opt/qgis', True)
q = QgsApplication([], False)
q.initQgis()


# uncomment out following to run as test env.

# Debug
#src = src = os.path.normpath(os.path.join( os.path.dirname(__file__),"testdata","Cells Null Coordinates.gas"))   #"Cells Null Coordinates.gas" #demo 2 shapes plus and open square.gas"
#createScratchLayer(src,'HIERARCHICAL')

# Layer 1: Create a Geopackage layer
#src = os.path.normpath(os.path.join( os.path.dirname(__file__),"testdata","demo_ManyLegendCombos.gas"))
#createGpkgLayer(src,'HIERARCHICAL')

# Layer 2: Create a Geopackage layer
src = os.path.normpath(os.path.join( os.path.dirname(__file__),"testdata","demo2.gas"))
createGpkgLayer(src,'LINEAR')


# Layer 3: Create as a Memory-Only Layer
#src = os.path.normpath(os.path.join( os.path.dirname(__file__),"testdata","GIS.SMALL.gas"))
#createGpkgLayer(src,'LINEAR')


# debug add same again
#src = os.path.normpath(os.path.join( os.path.dirname(__file__),"testdata","REE.gas"))
#createScratchLayer(src,'LINEAR')

print("Writing Completed, exiting...")